#include <cassert>
#include <iomanip>

#include "CoinPragma.hpp"
#include "OsiClpSolverInterface.hpp"

void checkDualObjValue(OsiClpSolverInterface si){
  int numCols = si.getNumCols(); // n
  int numRows = si.getNumRows(); // m
  double dual_obj_val = 0;

  // pi * b
  for (int i = 0; i < numRows; i++){
    if (si.getRowUpper()[i] < 1000000)
      dual_obj_val += si.getRowUpper()[i] * si.getRowPrice()[i];
    else
      dual_obj_val += si.getRowLower()[i] * si.getRowPrice()[i];
  }
  
  // dj * lb/ub
  for (int i = 0; i < numCols; i++)
  {
    if (si.getReducedCost()[i] < -1e-5)
        dual_obj_val += si.getColUpper()[i] * si.getReducedCost()[i];
    else if (si.getReducedCost()[i] > 1e-5)
        dual_obj_val += si.getColLower()[i] * si.getReducedCost()[i];
  }

  if (si.isProvenOptimal())
    printf("CLP: isProvenOptimal\n");
  else if (si.isIterationLimitReached())
    printf("CLP: isIterationLimitReached\n");
  else if (si.isProvenPrimalInfeasible())
    printf("CLP: isProvenPrimalInfeasible\n");

  printf("Dual objective: %.10f\n", dual_obj_val);
  printf(" getObjValue(): %.10f\n", si.getObjValue());

  if (fabs(dual_obj_val - si.getObjValue()) <= 0.0001)
    printf("Values matches!\n\n");
  else
    printf("WARNING: Values mismatches!\n\n");

}

int main(int argc, const char *argv[])
{
  OsiClpSolverInterface si;
  int numberIterations, numCols, numRows;

  std::string mpsFileName;
  if (argc >= 2) {
    mpsFileName = argv[1];

    int numMpsReadErrors = si.readMps(mpsFileName.c_str(), "");
   
    if (numMpsReadErrors != 0) {
      printf("%d errors reading MPS file\n", numMpsReadErrors);
      return numMpsReadErrors;
    }
  } else {
    printf("Need mps file\n");
    return -1;
  }

  // Setup OsiClp options
  si.setupForRepeatedUse(0, 0);
  int options = si.getModelPtr()->specialOptions();
  si.getModelPtr()->setSpecialOptions((options|32));
  si.getModelPtr()->setPerturbation(50);
  si.setCleanupScaling(1);

  // Only dual simplex should be used and no presolve
  si.setHintParam(OsiDoReducePrint);
  si.messageHandler()->setLogLevel(0);
  si.setHintParam(OsiDoScale,false,OsiHintDo);
  si.setHintParam(OsiDoPresolveInInitial, false, OsiHintDo);
  si.setHintParam(OsiDoPresolveInResolve, false, OsiHintDo);
  si.setHintParam(OsiDoDualInInitial, true, OsiHintDo);
  si.setHintParam(OsiDoDualInResolve, true, OsiHintDo);

  numCols = si.getNumCols(); // n
  numRows = si.getNumRows(); // m

  // Solve the "root node" LP
  printf("Solving the root LP...\n");
  si.initialSolve();
  numberIterations = si.getIterationCount();

  checkDualObjValue(si);

  // Solve the first child LP
  // Simulating strong branching, so we set a MaxNumIter = 40
  si.setColBounds(73, 0, 0);
  si.setIntParam(OsiMaxNumIteration, 40);
  printf("Solving the first child LP...\n");
  si.initialSolve();
  numberIterations = si.getIterationCount();
  
  checkDualObjValue(si);

  si.setupForRepeatedUse(2, 0);
  // Solve the second child LP
  // Simulating strong branching, so we set a MaxNumIter = 40
  si.setIntParam(OsiMaxNumIteration, 40);
  si.setColBounds(73, 1, 1);
  printf("Solving the second child LP...\n");
  si.initialSolve();
  numberIterations = si.getIterationCount();
  
  checkDualObjValue(si);

  // Problematic Ray
  si.readMps("lp_farkas_proof.mps", "");

  // Load the basis that led to the problematic ray
  si.getModelPtr()->readBasis("basis_ray.mps");

  si.setIntParam(OsiMaxNumIteration, 99999999);
  si.initialSolve();

  numberIterations = si.getIterationCount();
  
  checkDualObjValue(si);

  if (si.isProvenPrimalInfeasible()){
    double *ray = si.getDualRays(1, false)[0];
    double *rayA = (double *)malloc(sizeof(double) * numCols);

    const CoinPackedMatrix *A = si.getMatrixByCol();
         
    A->transposeTimes(ray, rayA);
    double farkasProof = 0;

    for (int i = 0; i < numRows; i++){
        farkasProof -= si.getRightHandSide()[i] * ray[i];
    }

    for (int j = 0; j < numCols; j++){
        if (rayA[j] > 1e-5){
          farkasProof += si.getColLower()[j] * rayA[j];
        } else if (rayA[j] < -1e-5) {
          farkasProof += si.getColUpper()[j] * rayA[j];
        }
    }

    if (farkasProof < -1e-5){
        printf("Farkas proof: rAx - rb > 0 fails!\n");
        printf("rAx - rb: %.10f\n", farkasProof);
    }

    free(ray);
    free(rayA); 
  }

  return 0;
}
